# Gridlock - a collection of grid-based strategy games
# Copyright (c) 2009 Brian Nenninger (brian@dozingcatsoftware.com)
# Gridlock comes with ABSOLUTELY NO WARRANTY, and is released under version 3 of the GNU GPL.

import GridUtils
from Game import GameState
import random

PAWN = 1
KNIGHT = 2
BISHOP = 3
ROOK = 4
QUEEN = 5
KING = 6

# if a piece is moved from one of the keys, the corresponding can_castle index(es) should be made False
CASTLE_LOSSES = {(0,0): (1,), (7,0): (0,), (4,0): (0,1), (0,7): (3,), (7,7): (2,), (4,7): (2,3),}
# squares that must be empty to castle
CASTLE_EMPTY_SQUARES = ( ((5,0),(6,0)), ((1,0),(2,0),(3,0)), ((5,7),(6,7)), ((1,7),(2,7),(3,7)) )
# squares that must be not threatened to castle (not including king start or end)
CASTLE_SAFE_SQUARES = ((5,0), (3,0), (5,7), (3,7))
# king/rook to/from
CASTLE_POSITIONS = ( ((4,0),(6,0),(7,0),(5,0)), ((4,0),(2,0),(0,0),(3,0)),
                        ((4,7),(6,7),(7,7),(5,7)), ((4,7),(2,7),(0,7),(3,7)) )

class ChessGameState (GameState):
    def reset(self):
        GameState.reset(self)
        # set up initial pieces, store castle and en passant states in extra_info
        firstrow1 = firstrow2 = [ROOK, KNIGHT, BISHOP, QUEEN, KING, BISHOP, KNIGHT, ROOK]
        self.extra_state = {'enpassant': -1, 'can_castle': [True, True, True, True]} # white K, white Q, black K, white Q
        # if king's corner variant, randomize starting position (with king in corner)
        if self.game_params.get('kingsCorner'):
            firstrow1[7] = KING
            # put bishops on opposite colors and randomize remainder
            b1 = random.randint(0,3)*2
            b2 = random.randint(0,2)*2+1
            firstrow1[b1] = firstrow1[b2] = BISHOP
            rest = [ROOK, ROOK, KNIGHT, KNIGHT, QUEEN]
            random.shuffle(rest)
            indexes = [i for i in range(7) if i!=b1 and i!=b2]
            for i,piece in enumerate(rest):
                firstrow1[indexes[i]] = rest[i]
            # black is reversed sequence of white, no castling
            firstrow2 = firstrow1[::-1]
            self.extra_state['can_castle'] = [False, False, False, False]
            
        for i in xrange(8):
            self.grid[i,0] =  firstrow1[i]
            self.grid[i,7] = -firstrow2[i]
            # look for variant where pawns are randomly placed on 2nd and 3rd rows
            pawncol = random.randint(1,2) if self.game_params.get('random23Pawns') else 1
            self.grid[i,pawncol] =  PAWN
            self.grid[i,7-pawncol] = -PAWN

    def copy_extra_state(self):
        return {'enpassant': self.extra_state['enpassant'], 'can_castle': self.extra_state['can_castle'][:]}
        
    def successors(self):
        # make sure the move doesn't expose the king
        for move,nextstate in self.successors_ignoring_check():
            if not is_king_threatened(nextstate.grid, self.current_player):
                yield (move,nextstate)
        
    def successors_ignoring_check(self):
        for pos,val in enumerate_player_pieces(self.grid, self.current_player):
            for npos in valid_chess_moves(pos, val, self.grid):
                nextgrid = self.grid.copy()
                nextgrid[pos] = 0
                # check for pawn promotion
                # TODO: allow promoting to other than queen
                if (val==PAWN and npos[1]==7) or (val==-PAWN and npos[1]==0):
                    nextgrid[npos] = QUEEN if val>0 else -QUEEN
                else:
                    nextgrid[npos] = val
                nextstate = self.next_state_with_grid(nextgrid)
                # copy/update castling state
                for ci in CASTLE_LOSSES.get(pos, ()):
                    nextstate.extra_state['can_castle'][ci] = False
                # update en passant state (store column if pawn moved 2 squares)
                if abs(val)==PAWN and abs(pos[1]-npos[1])==2:
                    nextstate.extra_state['enpassant'] = npos[0]
                yield ( (pos,npos), nextstate )
                
        # check for castling, king can't be in check
        if not is_king_threatened(self.grid, self.current_player):
            castleindexes, kingrow, pmul = ((0,1),0,1) if self.current_player==1 else ((2,3),7,-1)
            for ci in castleindexes:
                if self.extra_state['can_castle'][ci]:
                    # make sure intermediate squares are empty, king can't move through threatened square
                    if all(self.grid[ce]==0 for ce in CASTLE_EMPTY_SQUARES[ci]):
                        if not is_position_attackable(self.grid, CASTLE_SAFE_SQUARES[ci], self.next_player()):
                            # do the castle, clearing castling flags
                            nextgrid = self.grid.copy()
                            cps = CASTLE_POSITIONS[ci]
                            nextgrid[cps[0]] = 0
                            nextgrid[cps[1]] = pmul*KING
                            nextgrid[cps[2]] = 0
                            nextgrid[cps[3]] = pmul*ROOK
                            nextstate = self.next_state_with_grid(nextgrid)
                            for ci2 in castleindexes: nextstate.extra_state['can_castle'][ci2] = False
                            yield ( (cps[0],cps[1]), nextstate )
                        
        # check for en passant capture 
        if self.extra_state['enpassant']>=0:
            # pawns on same row can capture (assume epcapturerow is empty, otherwise enpassant wouldn't be >0)
            eprow, epdestrow, eppiece = (4,5,PAWN) if self.current_player==1 else (3,2,-PAWN)
            for eppos in ( (self.extra_state['enpassant']-1,eprow), (self.extra_state['enpassant']+1,eprow) ):
                if self.valid_coords(eppos) and self.grid[eppos]==eppiece:
                    nextgrid = self.grid.copy()
                    nextgrid[eppos] = 0
                    nextgrid[self.extra_state['enpassant'],eprow] = 0
                    nextgrid[self.extra_state['enpassant'],epdestrow] = eppiece
                    yield ( (eppos,(self.extra_state['enpassant'],epdestrow)), self.next_state_with_grid(nextgrid) )
    
    # don't need to override is_game_over because moves that leave the king exposed aren't valid
        
    def winning_player(self):
        # game is over, so current player can't move. If king is attackable he loses, if not it's a stalemate
        if is_king_threatened(self.grid, self.current_player):
            return self.next_player()
        return 0
        
    def has_scores(self):
        return False
        
    def animations_for_move(self, move, newstate):
        import PGPieceAnimation
        # this handles the move and possible capture at destination
        animations = PGPieceAnimation.move_with_capture_animations(move[0], move[1])
        # find other changed positions, which will be castling or en passant capture
        changed = [c for c in self.itercoords() if c not in move and self.grid[c]!=newstate.grid[c]]
        if len(changed)==2:
            # castle
            frompos = changed[0] if self.grid[changed[0]]!=0 else changed[1]
            topos = changed[0] if self.grid[changed[0]]==0 else changed[1]
            animations.append(PGPieceAnimation.move_animation(frompos, topos))
        elif len(changed)==1:
            # en passant capture
            animations.append(PGPieceAnimation.remove_animation(changed[0]))
        return PGPieceAnimation.AnimationGroup(animations)
        

def enumerate_player_pieces(grid, pnum):
    for pos,val in GridUtils.enumerate_all(grid):
        if (pnum==1 and val>0) or (pnum==2 and val<0):
            yield (pos,val)
                    
def valid_chess_moves(pos, piece, grid, attack_only=False):
    def is_friendly(p):
        return (piece>0 and p>0) or (piece<0 and p<0)
        
    def is_enemy(p):
        return (piece>0 and p<0) or (piece<0 and p>0)
        
    def moves_in_directions(*dirs):
        for d in dirs:
            for npos,val in GridUtils.enumerate_values_with_direction(grid, pos, d):
                # we can move here unless it's occupied by a friendly piece, any piece will stop further movement
                if not is_friendly(val): yield npos
                if val!=0: break
                    
    piecetype = abs(piece)
    if piecetype==PAWN:
        (dr, homerow) = (1,1) if piece>0 else (-1,6)
        # look forward 1 square, 2 if on home row
        # assuming pawn can never be on last row because it would promote
        # check for empty spaces 1 ahead, and 2 if on home row
        # if attack_only is true, ignore non-capturing forward move
        if not attack_only:
            forward1 = (pos[0], pos[1]+dr)
            if grid[forward1]==0: 
                yield forward1
                if pos[1]==homerow:
                    forward2 = (forward1[0], forward1[1]+dr)
                    if grid[forward2]==0: yield forward2
        # check for diagonal captures (if attack_only is true, target can be empty instead of enemy)
        for d in ((1,dr),(-1,dr)):
            npos = GridUtils.add_tuples(pos,d)
            if GridUtils.valid_coords(grid, npos) and ((is_enemy(grid[npos]) or (attack_only and grid[npos]==0))):
                yield npos
                
    elif piecetype==KNIGHT:
        # check all 8 directions, any non-friendly space is valid
        for d in ((1,2),(2,1),(-1,2),(-2,1),(1,-2),(2,-1),(-1,-2),(-2,-1)):
            npos = GridUtils.add_tuples(pos,d)
            if GridUtils.valid_coords(grid, npos) and not is_friendly(grid[npos]):
                yield npos
                
    # bishop/rook/queen use moves_in_directions helper
    elif piecetype==BISHOP:
        for npos in moves_in_directions((1,1),(-1,1),(1,-1),(-1,-1)): yield npos
    elif piecetype==ROOK:
        for npos in moves_in_directions((1,0),(-1,0),(0,1),(0,-1)): yield npos
    elif piecetype==QUEEN:
        for npos in moves_in_directions((1,0),(-1,0),(0,1),(0,-1),(1,1),(-1,1),(1,-1),(-1,-1)): yield npos
    elif piecetype==KING:
        # neighbors of radius 1
        for npos,nval in GridUtils.enumerate_neighbors(grid, pos):
            if not is_friendly(nval): yield npos

def is_position_attackable(grid, pos, pnum):
    for p,val in enumerate_player_pieces(grid, pnum):
        if pos in valid_chess_moves(p, val, grid, attack_only=True):
            return True
    return False
    
def is_king_threatened(grid, pnum):
    kpos = GridUtils.first_coord_with_value(grid, +KING if pnum==1 else -KING)
    if not kpos: return False
    return is_position_attackable(grid, kpos, 3-pnum)
            

from GenericAI import GenericAI

weights = {PAWN: 100, KNIGHT: 325, BISHOP: 325, ROOK: 500, QUEEN: 900}

class ChessAI (GenericAI):
    def game_utility(self, game):
        utility = 0
        for val in GridUtils.itervalues(game.grid):
            if val!=0:
                absval = abs(val)
                if absval in weights:
                    utility += (weights[absval] if val>0 else -weights[absval])
        if game.current_player==2: utility = -utility
        return utility
    

